/***************************************************************************
 *   Copyright (C) 2005 by Xin Hu   *
 *      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/**********************************************************************
 * panel quadrature of 1/R or e^{-jkR}/R
**********************************************************************/
#ifndef CALCPQUADRATURE_H_
#define CALCPQUADRATURE_H_

#include "../const_def.h"

template <class T1, class T2>
class calcPQuadrature
{
public:
  calcPQuadrature(void) {};
  void init(const complex<double> K_num, const size_t n=DEFAULT_NUM_GQ_PTS);
  void operator() (panel<T1> src_p, point3D<T2> eval,complex<double>& dlp);

private:
  void setupPanelLocalCoordSystem(panel<T1>&, const point3D<T2> &,point3D<T1 > &);
  complex<double> quad_panel(panel<T1>& src_p, const point3D<T1 >& eval, const int numQuadPts);
  complex<double> K_;
  int numQuadPts;
};


template <class T1, class T2>
void calcPQuadrature<T1, T2>::init (
  const complex<double> K_num, const size_t n)
{
  K_=K_num;
  numQuadPts=n;
}


template<class T1, class T2>
void calcPQuadrature<T1, T2>::
operator()(panel<T1> src_panel, point3D<T2> evalPt, complex<double>& dlp)
{

  point3D<T1 > evalPt_new(evalPt.x(),evalPt.y(),evalPt.z() );
  setupPanelLocalCoordSystem(src_panel, evalPt, evalPt_new);

  dlp=quad_panel(src_panel, evalPt_new, numQuadPts);

}

template<class T1, class T2>
void calcPQuadrature<T1, T2>::
setupPanelLocalCoordSystem (panel<T1>& src_panel, const point3D<T2>& evalPnt,
                            point3D<T1>& evalPnt_new)
{

  // 1) find the projection of evaluation point on the panel
  point3D<T1> evalPntProjection(evalPnt.x(), evalPnt.y(), evalPnt.z());
  evalPntProjection.transferGlobalToLocalCoord(src_panel.centroid(),
      src_panel.tangent1(),
      src_panel.tangent2(),
      src_panel.normal());

  T1 h = evalPntProjection.z();
  evalPntProjection = point3D<T1 >(evalPntProjection.x(), evalPntProjection.y(), 0.);

  // 2) shift back to the global coord system to be
  // compatible with all verteices which are in the globl coord.
  evalPntProjection.transferLocalToGlobalCoord(
    src_panel.centroid(),
    src_panel.tangent1(),
    src_panel.tangent2(),
    src_panel.normal() );

  // 3) use projection point as origin of the panel local coordinate system
  src_panel.shiftToLocalCoord(evalPntProjection);
  //evalPnt_new = point3D<complex<double> >(0., 0., h);
  evalPnt_new.transferGlobalToLocalCoord(evalPntProjection, src_panel.tangent1(),
                                         src_panel.tangent2(), src_panel.normal());
}

/**********************************************************************
  * panel quadrature
**********************************************************************/
template<class T1, class T2>
complex<double> calcPQuadrature<T1, T2> ::quad_panel(panel<T1>& src_p,
    const point3D<T1>& eval, const int numQuadPts)
{
  vector<point3D<T1> > quad_pts;
  vector<T1> quad_weights;
  T1 R;
  complex<double> val=0.;

  src_p.panel_quadrature(numQuadPts, quad_pts, quad_weights);


  for (int i=0; i<quad_pts.size(); i++)
  {
    R=length(eval-quad_pts[i]);
    val=val+exp(-1.*IMAG* K_*R)/R*quad_weights[i];
  }

  return val;

}

#endif
