/***************************************************************************
 *   Copyright (C) 2005 by Xin Hu   *
 *   xinhu@mit.edu   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "sysformation.h"
#include "filament.h"
#include "const_def.h"
#include "gmres.h"
#include "../pfft/cmat.h"
#include "../pfft/vec.h"


sysFormation::sysFormation(Structure& structure,
                           const LayerEnv& layerenv,
                           const SimulationType simuType,
                           const bool debug): condStruc(structure),
    layers(layerenv), simuType_(simuType), EMQS_flag(false), dipole_flag(false),
    debug_flag(debug)
{

  numPanels=condStruc.num_panels();
  numFilaments=condStruc.num_filaments();
  numLayers_=layers.num_layers();


  if ((simuType_==EMQS) && (numLayers_==1))
  {
    EMQS_flag=true;
    dipole_flag=false;
  }
  else if ((simuType_==EMQS) && (numLayers_>1))
  {
    EMQS_flag=true;
    dipole_flag=true;
  }
  else if ((simuType_==FULLWAVE) && (numLayers_==1))
  {
    EMQS_flag=false;
    dipole_flag=false;
  }
  else if ((simuType_==FULLWAVE) && (numLayers_>1))
  {
    EMQS_flag=false;
    dipole_flag=true;
  }
  else
    cout<<"unknown options "<<endl;

  //initialized inductance and charge calculations
  inductance.init(EMQS_flag, dipole_flag,layers.get_epir_topmostlayer()*EPI0);
  chg.init(EMQS_flag, dipole_flag, condStruc, layers.get_epir_topmostlayer()*EPI0);



  //if EMQS mode is used and there is no substrate, precompute the L and P
  //matrices which are freq-independent.
  if ((EMQS_flag==true) && (dipole_flag==false))
  {
// TAREK MOSELHY
 //   fill_LMat();
 //   fill_PMat();
  }
  //R and M matrices should only be computed as well
  fill_RMat();
  assemble_M_mat();
}


/*******************************************************************
* Assemble M matrix for nodal analysis
*******************************************************************/
void sysFormation::assemble_M_mat()
{
  pfft::SpRowMat<double> Mf, Mfs, Mp, Mps, Mrs;
  sparse sVs;
  pfft::SpVec<double> Vs_temp;

  int err;

  const int Mf_nRow=(condStruc.mesh_mats).Mf_ptr.numRow();
  const int Mf_nCol=(condStruc.mesh_mats).Mf_ptr.numCol();
  const int Mfs_nRow=(condStruc.mesh_mats).Mfs_ptr.numRow();
  const int Mfs_nCol=(condStruc.mesh_mats).Mfs_ptr.numCol();
  const int Mp_nRow=(condStruc.mesh_mats).Mp_ptr.numRow();
  const int Mp_nCol=(condStruc.mesh_mats).Mp_ptr.numCol();
  const int Mps_nRow=(condStruc.mesh_mats).Mps_ptr.numRow();
  const int Mps_nCol=(condStruc.mesh_mats).Mps_ptr.numCol();
  const int Mrs_nRow=(condStruc.mesh_mats).Mrs_ptr.numRow();
  const int Mrs_nCol=(condStruc.mesh_mats).Mrs_ptr.numCol();
  const int Vs_nRow=(condStruc.mesh_mats).Vs_ptr.size();
 
 
 // const int M_nRow=Mf_nRow+Mfs_nRow+Mrs_nRow+Mp_nRow-Vs_nRow;
  const int M_nRow=Mf_nRow+Mfs_nRow+Mrs_nRow+Mp_nRow;
  const int M_nCol=Mfs_nCol+Mrs_nCol+Mp_nCol;


  
  //copy into M_mat
  /*
  M=pfft::SpRowMat<double>(M_nRow, numFilaments+Mrs_nRow+numPanels);
  insert_subMatrix(M, 0, 0, (condStruc.mesh_mats).Mf_ptr);
  insert_subMatrix(M, Mf_nRow,0, (condStruc.mesh_mats).Mfs_ptr);
  insert_subMatrix(M, Mf_nRow+Mfs_nRow, numFilaments, (condStruc.mesh_mats).Mrs_ptr);
  insert_subMatrix(M, Mf_nRow, numFilaments+Mrs_nRow, (condStruc.mesh_mats).Mps_ptr);
  insert_subMatrix(M, Mf_nRow+Mfs_nRow+Mrs_nRow, numFilaments+Mrs_nRow, (condStruc.mesh_mats).Mp_ptr);
  */
  
  //copy into M_mat
  M=pfft::SpRowMat<double>(M_nRow, numFilaments+numPanels+Mrs_nRow);
  insert_subMatrix(M, 0, 0, (condStruc.mesh_mats).Mf_ptr);
  insert_subMatrix(M, Mf_nRow,0, (condStruc.mesh_mats).Mfs_ptr);
  insert_subMatrix(M, Mf_nRow+Mfs_nRow, numFilaments+numPanels, (condStruc.mesh_mats).Mrs_ptr);
  insert_subMatrix(M, Mf_nRow, numFilaments, (condStruc.mesh_mats).Mps_ptr);
  insert_subMatrix(M, Mf_nRow+Mfs_nRow+Mrs_nRow, numFilaments, (condStruc.mesh_mats).Mp_ptr);
  
  debug_flag = 1;
// TAREK MOSELHY
  if(debug_flag)
    output_to_file_real("M.mat", M);
  port_location_ = Mf_nRow+Mfs_nRow;

}


/*******************************************************************
* filling R, L and C matrices
*******************************************************************/
void sysFormation::fill_RMat()
{
  double R;
  filament<double> *f;
  int i;


  RMat=pfft::SpRowMat<double>(numFilaments, numFilaments);
  for (f=condStruc.list_filaments.initialize(), i=0; f!=NULL;
       f=condStruc.list_filaments.iterate())
  {
    R=f->filLength()/(f->sigma()*f->filCrossA());
    RMat.insertElement(i,i,R);
    i++;
  }
  debug_flag = 1;
  if (debug_flag)
    output_to_file_real("R.mat", RMat);

}


void sysFormation::fill_LMat()
{
  const filament<double> *f1, *f2;
  int i,j;

  LMat=pfft::SpRowMat<complex<double> >(numFilaments, numFilaments);
  for (f1=condStruc.list_filaments.initialize(), i=0; f1!=NULL;
       f1=condStruc.list_filaments.iterate())
  {
    for (f2=condStruc.list_filaments.initialize(2),j=0; f2!=NULL;
         f2=condStruc.list_filaments.iterate(2))
    {
      inductance.fill_L( LMat, i, j, f1, f2);
      j++;
    }
    i++;
  }
	debug_flag = 1;
  if (debug_flag)
    output_to_file_complex("L.mat", LMat);
}


void sysFormation::fill_PMat()
{
  panel<double> *p1, *p2;
  int i,j;

  PMat=pfft::SpRowMat<complex<double> >(numPanels, numPanels);
  for (p1=condStruc.list_panels.initialize(), i=0; p1!=NULL;
       p1=condStruc.list_panels.iterate())
  {
    for (p2=condStruc.list_panels.initialize(2), j=0; p2!=NULL;
         p2=condStruc.list_panels.iterate(2))
    {
      chg.fill_p(PMat, i,j, p1, p2);
      j++;
    }
    i++;
  }
  debug_flag = 1;
  if (debug_flag)
    output_to_file_complex("P.mat", PMat);
}


/*******************************************************************
* Setting up system by providing freq info
*******************************************************************/
void sysFormation::setup_system(const Dipoles& dipoles,
                                const double freq)
{
  setup_freq_info(freq);
  setup_loads(zLoads);
  //if requiring full-wave or there is a substrate
  //then compute L and P matrices for each freq point
  if  (( dipole_flag==true)||(EMQS_flag==false))
  {
    inductance.set_dipoles(dipoles, freq);
    chg.set_dipoles(dipoles, freq);
    fill_PMat();
    cout<<"finished PMat"<<endl;
    fill_LMat();
    cout<<"finished LMat"<<endl;
  }
  setup_ZEM(zLoads);
  cout<<"finshed setting up zem"<<endl;
  M_mult_ZEM_mult_MT (mZEMmt, ZEM, M);
  cout<<"finshed multiplying MZemMT"<<endl;
  setup_RHS(zLoads);
  cout<<"System matrix of size "<<mZEMmt.numRow()<<
  " by "<<mZEMmt.numCol()<<" has been setup successfully"<<endl;

}

void sysFormation::setup_freq_info (double freq)
{
  omega=2*PI*freq;
  k0=layers.get_k_topmostLayer();
}

//void sysFormation::setup_loads(vector<complex<double> >& zLoads)
//{
//  Ext_Node_Pair *ext;
//  Resistance *resist;;
//  complex<double> Z;
//
//  //condStruc.find_load_resistance();
//  //obtaining loads by examining the resistance attached to
//  //each port
//  for (resist=condStruc.list_resistances.initialize(); resist!=NULL;
//      resist=condStruc.list_resistances.iterate())
// {
//    //rList= ext->get_resistance_list();
//    //assuming all z values are R; no C or L as loads
//    Z=resist->get_r_value();
//    zLoads.push_back(Z);
//  }
//}

void sysFormation::setup_loads(vector<complex<double> >& zLoads)
{
  Ext_Node_Pair *ext;
  vector<Resistance*> rList;
  complex<double> Z;

  condStruc.find_load_resistance();
  //obtaining loads by examining the resistance attached to
  //each port
  for (ext=condStruc.list_ext_node_pairs.initialize(); ext!=NULL;
       ext=condStruc.list_ext_node_pairs.iterate())
  {
    rList= ext->get_resistance_list();
    Z=0;
    //assuming all z values are R; no C or L as loads
    for (int i=0; i<rList.size(); i++)
      Z=Z+rList[i]->get_r_value();
    zLoads.push_back(Z);
  }
}


void sysFormation::setup_ZEM(const vector<complex<double> >& zLoads)
{
  pfft::SpRowMat<std::complex<double> > sL, Ps, loadMat, RsL;
  sL=pfft::SpRowMat<complex<double> >(numFilaments, numFilaments);
  Ps=pfft::SpRowMat<complex<double> >(numPanels, numPanels);
  ZEM=pfft::SpColMat<std::complex<double> >(M.numCol(),M.numCol());
  RsL= pfft::SpRowMat<std::complex<double> >(numFilaments, numFilaments);

  int numLoads=M.numCol()-(sL.numRow()+Ps.numRow());
  loadMat=pfft::SpRowMat<std::complex<double> >(numLoads, numLoads);


  matMultScalar(sL, LMat, IMAG*omega);
  matDivScalar(Ps, PMat, IMAG*omega);
  matAdd(RsL, sL, RMat);



  for (int i=0; i<numLoads; i++)
    loadMat.insertElement(i,i,zLoads[i]);
  insert_subMatrix_spColMat(ZEM,0,0,RsL);
  insert_subMatrix_spColMat(ZEM, sL.numRow(), sL.numRow(),
                            loadMat);
  insert_subMatrix_spColMat(ZEM, sL.numRow()+numLoads,
                            sL.numRow()+numLoads, Ps);

  debug_flag = 1;
  //if (debug_flag)
  //  output_to_file_complex("P.mat", PMat);

     FILE *fp, *fp2;
     fp = fopen ("ZEM_real.mat", "w");
     fp2 = fopen ("ZEM_imag.mat", "w");
     for (size_t col=0; col<ZEM.numCol(); col++)
     {
   	  for (size_t row=0; row<ZEM[col].size(); row++)
   	  {
   		  fprintf(fp, "%i %i %e ", ZEM[col].index(row)+1, col+1, real(ZEM[col].value(row) ));
   		  fprintf(fp2, "%i %i %e ", ZEM[col].index(row)+1, col+1,  imag(ZEM[col].value(row)));
   		  fprintf(fp, "\n");
   		  fprintf(fp2, "\n");
   	  }
     }
     fclose(fp);
     fclose(fp2);

}

void sysFormation::setup_RHS(const vector<complex<double> >& zLoads)
{
/*  int i, startIndex;

  startIndex=(condStruc.mesh_mats).Mf_ptr->get_n_lines()+
             (condStruc.mesh_mats).Mfs_ptr->get_n_lines();
  changeTo_sparse_src(*condStruc.mesh_mats.Vs_ptr, Vs, startIndex, M.numRow(),
                      "Vs.mat", "Vs");

  Vm.resize(M.numRow());
  for (i=0; i<M.numRow(); i++)
    Vm[i]=0.;
  for (i=0; i<Vs.size(); i++)
    Vm[Vs.index(i)]=zLoads[i];

 // debug_flag = 1;
  if(debug_flag)
  {
    FILE* vm_file;
    vm_file=fopen("Vm.mat", "w");
    for (i=0; i<Vm.size(); i++)
      fprintf(vm_file, "%i %e %e\n ", i+1, real(Vm[i]), imag(Vm[i]));
    fclose(vm_file);
  }
  */
}


void sysFormation::M_mult_ZEM_mult_MT (pfft::SpColMat<std::complex<double> >& mZEMmt,
                                       const pfft::SpColMat<std::complex<double> >& ZEM,
                                       const pfft::SpRowMat<double>& Mx)
{
/*
//     complex<double> fullVecAns[ZEM.numCol()];
//     complex<double> fullVecAns2[M.numRow()];
//     double fullVec[M.numCol()];
//     for (int i=0; i<M.numRow(); i++)
//     {
//       convert_to_full_vec(fullVec, M[i], M.numCol()); //M is checked
//       vecMultMat_fv(fullVecAns, fullVec, ZEM, ZERO);  //ZEM checked
//       vecMultMatT_fv(fullVecAns2, fullVecAns, M, ZERO);
//       for (int j=0; j<M.numRow(); j++)
//         mZEMmt.insertElement(i,j,fullVecAns2[j]);
//     }


  int row, col;
  complex<double> ZEMcol[ZEM.numCol()];
  complex<double> mZEMrow[Mx.numRow()];
  complex<double> mZEMcol[Mx.numRow()];
  complex<double> mZEMmtRow[Mx.numRow()];
  pfft::SpRowMat<std::complex<double> > mZEM;

  mZEMmt=pfft::SpColMat<complex<double> >(Mx.numRow(), Mx.numRow());
  mZEM=pfft::SpRowMat<complex<double> >(Mx.numRow(), ZEM.numCol());


  for (col=0; col<ZEM.numCol(); col++)
  {
    convert_to_full_vec(ZEMcol, ZEM[col], ZEM.numRow());
    matMultVec_fv(mZEMcol, Mx, ZEMcol);
    for (row=0; row<Mx.numRow(); row++)
      	mZEM.insertElement(row,col,mZEMcol[row]);
     // mZEM[row][col]=mZEMcol[row];
  }

//=ZEM[row].value(col);
  for (row=0; row<Mx.numRow(); row++)
  {
    convert_to_full_vec(mZEMrow,mZEM[row], mZEM.numCol());
    matMultVec_fv(mZEMmtRow,Mx,mZEMrow);
   // matMultVec_fv(mZEMmtRow,M,mZEM[row]);
    for (col=0; col<Mx.numRow(); col++)
    {
      if (mZEMmtRow[col] !=0.)
        mZEMmt.insertElement(row,col,mZEMmtRow[col]);
    }
  }
 // debug_flag = 1;
  if (debug_flag)
    output_to_file_complex("mzemmt.mat", mZEMmt);
*/
}


void sysFormation::solve_system_direct(vector<complex<double> >& Im,
                                       vector<complex<double> >& Ib,
                                       complex<double>& adm,
                                       complex<double>& Rsource)
{

/*
  //superLU.setup(mZEMmt);
  //superLU.factor();
  cout<<"system has been factored"<<endl;
  //superLU.solve(Vm, Im);
	for (int i=0; i<Im.size(); i++)
		cout << Im[i] << endl;
  //cout<<"norm is "<<superLU.estimateCondNum_OneNorm()<<endl;


  complex<double> Ib_fv[M.numCol()];
  int i,j;
  //since M is in spRowMat format, hard to do matrix transpose
  pfft::SpColMat<double> M_col=pfft::SpColMat<double>(M.numCol(),M.numRow());
  for (i=0; i<M.numRow(); i++)
  {
    for (j=0; j < M[i].size(); j++)
    {
      M_col.insertElement(i,M[i].index(j),M[i].value(j));
    }
  }
  matTMultVec_fv(Ib_fv, M_col, Im);
  for (int i=0; i<M.numCol(); i++)
    Ib.push_back(Ib_fv[i]);

  if (Vs.index(0)>Im.size())
    cout<<"something wrong here"<<endl;
  else
  {
    adm=Im[Vs.index(0)];
  }
  Rsource=zLoads[0];
*/
}


/*******************************************************************
* misc. auxillary functions
*******************************************************************/
/*
void sysFormation::changeTo_sparse_src(Source_Vector &vs,
                                       pfft::SpVec<complex<double> >& VsVec,
                                       const int startIndexInM,
                                       const int numVecElements,
                                       char* sName,
                                       char* name)
{
  pfft::SpVecElement<complex<double> > elem;

  for (int i=0; i<(int)(vs.num_sources()/2); i++)
  {
    elem= pfft::SpVecElement<complex<double> >(startIndexInM+i,ONE);
    VsVec.push_back(elem);
  }

}
*/



/*****************************************************************/
// this can be moved to fastsub!! and a complete new file instead of systemFormation can be used

void sysFormation::Solver(const Dipoles& dipoles,
			  const double freq, const SimulationType simuType,
			  const FormulationType formulationType,
			  const SolverType solverTypeIn,
			  FILE* out_file)
{

  if (formulationType==FORMULATION_A) {
  	if (solverTypeIn==DIRECT_SOLVER) {
  		// call my direct
  	}
  	else if (solverTypeIn==pFFT) {
  		pfftDRIVER_A(dipoles, freq, simuType);
  	}
  } else {
  //formulationType==FORMULATION_C
  	if (solverTypeIn==DIRECT_SOLVER) {
  		// call Xin's direct
  	}
  	else if (solverTypeIn==pFFT) {
  		pfftDRIVER_C(dipoles, freq, simuType, out_file);
  	}
  }				  
}


///////////////////////

/********************************************************************************
**			pfftDRIVER_A --
**
*******************************************************************************/

void sysFormation::pfftDRIVER_A(const Dipoles& dipoles,
                                const double freq,
																const SimulationType simuType)
{
bool CONVERGE;


setup_M_Elements();

TNT::Vector<complex<double> > RHS(Mpfft.numRow());
TNT::Vector<complex<double> > x0(Mpfft.numRow());

setup_pfft_A(dipoles, freq, simuType);

// this is moved after the setup_pfft_A because you might need to calculate the preconditioner before
// setting up the right hand side and the initial guess
RHS = setup_RHS_InitialGuess();
// this is the initial guess and the solution
for (int i=0; i<Mpfft.numRow(); i++)
	x0[i]=0.;
//	x0[i]=RHS[i];

// the matrix vector product function must be devided into two (three) functions according to the
// vector matrix product
// 1. Direct matrix vector product
// 2. Matrix Vector product using formulation C
// 3. matrix vector product using formulation A
CONVERGE=gmres(this, x0, RHS,MpreCondMatMT, 100, 50, 1e-6);

		cout << x0 << endl;
}
//////////
/*************************************************************************
* setup pfft this function serves to setup the lements and call
* all the different pfft routines with the appropriate initialization
**************************************************************************/
void sysFormation::setup_pfft_A(const Dipoles& dipoles,
                                const double freq,
																const SimulationType simuType)
{

 // pfft::DyadicGreenFunction DyadicGreenFunc;
 // DyadicGreenFunc=pfft::GXXa;
	setup_freq_info(freq);
  outerOperatorList.push_back(pfft::D0_D0);
  innerOperatorList.push_back(pfft::D0_D0);
  outerOperatorListX.push_back(pfft::D_DX);
  innerOperatorListX.push_back(pfft::D_DX);
  outerOperatorListY.push_back(pfft::D_DY);
  innerOperatorListY.push_back(pfft::D_DY);





  ofstream fout("error.dat");
  // good idea to have an accuracy flag!!
	  size_t max_size = 1;
    size_t interpStencilSize =  max_size;
    size_t projectStencilSize = max_size;
    size_t directStencilSize = 2*max_size;
    cout << endl << "\t Calculate matrix vector product indirectly using pfft" << endl;


	if (ElementListX.size() > 0) {



		if (simuType==FULLWAVE) {
		calcA_XX=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0, dipole_flag,
																						dipoles, pfft::GXXc,FULLWAVE);
			 kernel_XX=DynamicGreenFunc(k0, dipoles, pfft::GXXc);
	  } else if (simuType==EMQS) {
		calcA_XX=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0, dipole_flag,
																						dipoles, pfft::GXXc,EMQS);
			 kernel_XX=DynamicGreenFunc(0., dipoles, pfft::GXXc);
    }
		kernel_XX.set_shift(get_shift(ElementListX));
    pfft_XX.PfftConstruct(ElementListX, ElementListX, kernel_XX, calcA_XX,
		 directStencilSize, projectStencilSize, interpStencilSize,
						pfft::TOEPLITZplusHANKEL);
  }

// enter only if dipole is true
	if (ElementListX.size() > 0) {

		if (simuType==FULLWAVE) {
		calcA_ddXX=CalcAfullwave3D(outerOperatorListX, innerOperatorListX, k0, dipole_flag,
			 dipoles, pfft::GXXa,FULLWAVE);
			 kernel_XX_MOD=DynamicGreenFunc(k0, dipoles, pfft::GXXa);   
	} else if (simuType==EMQS) {
		calcA_ddXX=CalcAfullwave3D(outerOperatorListX, innerOperatorListX, k0, dipole_flag,
			 dipoles, pfft::GXXa,EMQS);
			 kernel_XX_MOD=DynamicGreenFunc(0., dipoles, pfft::GXXa);
	}
		kernel_XX_MOD.set_shift(get_shift(ElementListX));


    pfft_XX_MOD.PfftConstruct(ElementListX, ElementListX, kernel_XX_MOD, calcA_ddXX,
		 directStencilSize, projectStencilSize, interpStencilSize, pfft::HANKEL);

  }

 // do the same as in the previous section
	if (ElementListY.size() > 0) {


		if (simuType==FULLWAVE) {
		calcA_YY=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0, dipole_flag,
			dipoles,  pfft::GXXc,FULLWAVE);
		 kernel_YY=DynamicGreenFunc(k0, dipoles, pfft::GXXc);
		} else if (simuType==EMQS) {
		calcA_YY=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0, dipole_flag,
			dipoles,  pfft::GXXc,EMQS);
		 kernel_YY=DynamicGreenFunc(0., dipoles, pfft::GXXc);
    }

		kernel_YY.set_shift(get_shift(ElementListY));
    pfft_YY.PfftConstruct(ElementListY, ElementListY, kernel_YY, calcA_YY,
		 directStencilSize, projectStencilSize, interpStencilSize, pfft::TOEPLITZplusHANKEL);
  }

	if (ElementListY.size() > 0) {


		if (simuType==FULLWAVE) {
		calcA_ddYY=CalcAfullwave3D(outerOperatorListY, innerOperatorListY, k0,
				dipole_flag, dipoles, pfft::GYYa,FULLWAVE);
		 kernel_YY_MOD=DynamicGreenFunc(k0, dipoles, pfft::GXXa);
		} else if (simuType==EMQS) {
		calcA_ddYY=CalcAfullwave3D(outerOperatorListY, innerOperatorListY, k0,
				dipole_flag, dipoles, pfft::GYYa,EMQS);
		 kernel_YY_MOD=DynamicGreenFunc(0., dipoles, pfft::GXXa);
     }
		kernel_YY_MOD.set_shift(get_shift(ElementListY));
    pfft_YY_MOD.PfftConstruct(ElementListY, ElementListY, kernel_YY_MOD, calcA_ddYY,
		 directStencilSize, projectStencilSize, interpStencilSize, pfft::HANKEL);
  }



	if (ElementListX.size() > 0 && ElementListY.size() > 0) {

		if (simuType==FULLWAVE) {
		calcA_XY=CalcAfullwave3D(outerOperatorListX, innerOperatorListY, k0,
				dipole_flag, dipoles, pfft::GXYa,FULLWAVE);
		 kernel_XY=DynamicGreenFunc(k0, dipoles, pfft::GXXa);
		} else if (simuType==EMQS) {
		calcA_XY=CalcAfullwave3D(outerOperatorListX, innerOperatorListY, k0,
				dipole_flag, dipoles, pfft::GXYa,EMQS);
		 kernel_XY=DynamicGreenFunc(0., dipoles, pfft::GXXa);
    }

		double minZ=min(get_shift(ElementListX),get_shift(ElementListY));
		kernel_XY.set_shift(minZ);
    pfft_XY.PfftConstruct(ElementListX, ElementListY, kernel_XY, calcA_XY,
		 directStencilSize, projectStencilSize, interpStencilSize, pfft::HANKEL);
  }


	if (ElementListY.size() > 0 && ElementListX.size() > 0) {

		if (simuType==FULLWAVE) {
		calcA_YX=CalcAfullwave3D(outerOperatorListY, innerOperatorListX, k0, dipole_flag,
				dipoles, pfft::GYXa,FULLWAVE);
		 kernel_YX=DynamicGreenFunc(k0, dipoles, pfft::GXXa);
		} else if (simuType==EMQS) {
		calcA_YX=CalcAfullwave3D(outerOperatorListY, innerOperatorListX, k0, dipole_flag,
				dipoles, pfft::GYXa,EMQS);
		 kernel_YX=DynamicGreenFunc(0., dipoles, pfft::GXXa);
    }
		double minZ=min(get_shift(ElementListY),get_shift(ElementListX));
		kernel_YX.set_shift(minZ);
    pfft_YX.PfftConstruct(ElementListY, ElementListX, kernel_YX, calcA_YX,
		 directStencilSize, projectStencilSize, interpStencilSize, pfft::HANKEL);
  }


	if (ElementListZ.size() > 0) {

		if (simuType==FULLWAVE) {
		calcA_ZZ=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0,
						dipole_flag, dipoles, pfft::GZZ,FULLWAVE);
		 kernel_ZZ=DynamicGreenFunc(k0, dipoles, pfft::GZZ);
		} else if (simuType==EMQS) {
		calcA_ZZ=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0,
				dipole_flag, dipoles, pfft::GZZ,EMQS);
		 kernel_ZZ=DynamicGreenFunc(0., dipoles, pfft::GZZ);
    }

		kernel_ZZ.set_shift(get_shift(ElementListZ));
    pfft_ZZ.PfftConstruct(ElementListZ, ElementListZ, kernel_ZZ, calcA_ZZ,
		 directStencilSize, projectStencilSize, interpStencilSize, pfft::TOEPLITZplusHANKEL);
  }

	if (ElementListP.size() > 0) {
		if (simuType==FULLWAVE) {
		calcP_VERT=CalcPfullwave3D(outerOperatorList, innerOperatorList, k0,
						dipole_flag, dipoles, pfft::GZZ_phi, simuType);
		 kernel_P=DynamicGreenFunc(k0, dipoles, pfft::GZZ_phi);
		} else if (simuType==EMQS) {
		calcP_VERT=CalcPfullwave3D(outerOperatorList, innerOperatorList, k0,
				dipole_flag, dipoles, pfft::GZZ_phi, simuType);
		 kernel_P=DynamicGreenFunc(0., dipoles, pfft::GZZ_phi);
    }

		kernel_P.set_shift(get_shift(ElementListP));
    pfft_P_H.PfftConstruct(ElementListP, ElementListP, kernel_P, calcP_VERT,
		 directStencilSize, projectStencilSize, interpStencilSize,
				pfft::TOEPLITZplusHANKEL);
  }

setup_PRECONDITIONER();
}


/////////
TNT::Vector<complex<double> > sysFormation::MatrixVectorProduct_A(
													TNT::Vector<complex<double> > RHS)
{


	TNT::Vector<complex<double> > AP_Vmpfft(RHS.size());
	TNT::Vector<complex<double> > IbX(ElementListX.size());
	TNT::Vector<complex<double> > IbY(ElementListY.size());
	TNT::Vector<complex<double> > IbZ(ElementListZ.size());
	TNT::Vector<complex<double> > IbP(ElementListP.size());

	TNT::Vector<complex<double> > Ib(MTpfft.numRow());

	matMultVec(Ib,MTpfft,RHS);

	for (int ix=0; ix<ElementListX.size(); ix++)
		IbX[ix]=Ib[ix];
	for (int iy=0; iy<ElementListY.size(); iy++)
		IbY[iy]=Ib[iy+ElementListX.size()];
	for (int iz=0; iz<ElementListZ.size(); iz++)
		IbZ[iz]=Ib[iz+ElementListX.size()+ElementListY.size()];
	for (int ip=0; ip<ElementListP.size(); ip++)
		IbP[ip]=Ib[ip+ElementListX.size()+ElementListY.size()+ElementListZ.size()];


	TNT::Vector<complex<double> > Vb(ElementListX.size()+ElementListY.size()
																	+ElementListZ.size()+ElementListP.size());

  TNT::Vector<complex<double> > y_XX(ElementListX.size());
  TNT::Vector<complex<double> > y_YY(ElementListY.size());
  TNT::Vector<complex<double> > y_XX_MOD(ElementListX.size());
  TNT::Vector<complex<double> > y_YY_MOD(ElementListY.size());
  TNT::Vector<complex<double> > y_XY(ElementListY.size());
  TNT::Vector<complex<double> > y_YX(ElementListX.size());
  TNT::Vector<complex<double> > y_ZZ(ElementListZ.size());
  TNT::Vector<complex<double> > y_P(ElementListP.size());


	if (ElementListP.size() > 0) {
    pfft_P_H.IEoperator(y_P, outerOperatorList[0], innerOperatorList[0],
		    IbP);
	}

	if (ElementListX.size() > 0) {
    pfft_XX.IEoperator(y_XX, outerOperatorList[0], innerOperatorList[0],
		    IbX);
			//cout << y_XX << endl;

    pfft_XX_MOD.IEoperator(y_XX_MOD, outerOperatorListX[0], innerOperatorListX[0],
		    IbX);
			//cout << y_XX_MOD << endl;
  }
	if (ElementListY.size() > 0) {
    pfft_YY.IEoperator(y_YY, outerOperatorList[0], innerOperatorList[0],
		    IbY);
	//cout << y_YY << endl;
    pfft_YY_MOD.IEoperator(y_YY_MOD, outerOperatorListY[0], innerOperatorListY[0],
		    IbY);
	//cout << y_YY_MOD << endl;
  }
	if (ElementListX.size() > 0 && ElementListY.size() > 0) {
    pfft_XY.IEoperator(y_XY, outerOperatorListX[0], innerOperatorListY[0],
		    IbX);
	//cout << y_XY << endl;
	}

	if (ElementListY.size() > 0 && ElementListX.size() > 0) {
    pfft_YX.IEoperator(y_YX, outerOperatorListY[0], innerOperatorListX[0],
		    IbY);
	//cout << y_YX << endl;
  }
	if (ElementListZ.size() > 0) {
    pfft_ZZ.IEoperator(y_ZZ, outerOperatorList[0], innerOperatorList[0],
		    IbZ);
	}

	for (int i=0; i<ElementListX.size(); i++) {
  	Vb[i]=(y_XX[i]+y_XX_MOD[i]+y_YX[i])*IMAG*omega*1e-7;
	}
	for (int i=0; i<ElementListY.size(); i++) {
  	Vb[ElementListX.size()+i]=
				(y_YY[i]+y_YY_MOD[i]+y_XY[i])*IMAG*omega*1e-7;
	}
	for (int i=0; i<ElementListZ.size(); i++) {
  	Vb[ElementListX.size()+ElementListY.size()+i]
				=(y_ZZ[i])*IMAG*omega*1e-7;
	}

	for (int i=0; i<ElementListP.size(); i++) {
  	Vb[ElementListX.size()+ElementListY.size()+ElementListZ.size()+i]
				=y_P[i]/(IMAG*omega*4.*M_PI*layers.get_epir_topmostlayer()*EPI0);
	}

	//cout << Vb << endl;
	matMultVec(AP_Vmpfft,Mpfft,Vb);

	TNT::Vector<complex<double> > VR(ElementListX.size()+ElementListY.size()+ElementListZ.size());
	TNT::Vector<complex<double> > Ibr(ElementListX.size()+ElementListY.size()+ElementListZ.size());
	TNT::Vector<complex<double> > VR_PADDED(MTpfft.numRow());
	for (int i=0; i<ElementListX.size()+ElementListY.size()+ElementListZ.size();i++)
		Ibr[i]=Ib[i];
		
		
		

		
		
	matMultVec(VR,RMat,Ibr);
	for (int i=0; i<ElementListX.size()+ElementListY.size()+ElementListZ.size();i++)
		VR_PADDED[i]=VR[i];
	for (int i=0; i<ElementListP.size();i++)
		VR_PADDED[i+ElementListX.size()+ElementListY.size()+ElementListZ.size()]=0.;


	matMultVecAdd(AP_Vmpfft,Mpfft,VR_PADDED);

	return AP_Vmpfft;

}






///////////////////////




/********************************************************************************
**			IterativeSolver --
** 
*******************************************************************************/

void sysFormation::pfftDRIVER_C(const Dipoles& dipoles,
                                const double freq,
				const SimulationType simuType,
				FILE* out_file)
  
{
  bool CONVERGE;
  setup_M_Elements();
  
  TNT::Vector<complex<double> > RHS(Mpfft.numRow());
  TNT::Vector<complex<double> > x0(Mpfft.numRow());
  
  setup_pfft_C(dipoles, freq, simuType);
  
  RHS = setup_RHS_InitialGuess();
  for (int i=0; i<Mpfft.numRow(); i++)
    x0[i]=0.;
  //	x0[i]=RHS[i];
  cout << "finished Preconditioner Setup" << endl;
  CONVERGE=gmres(this, x0, RHS,MpreCondMatMT, Mpfft.numRow(), 100, 1e-4);
  
  //cout << x0[port_location_] << endl;
  //cout << x0[Mpfft.numRow()-2] << endl;
  //cout << x0[Mpfft.numRow()-1] << endl;
  
  // cout << x0 << endl;
  //debug_flag = true;
  //if(debug_flag)
  //{
  //FILE* x0_file;
  out_file=fopen("data.mat", "a");
  //for (int i=0; i<x0.size(); i++)
  complex<double> Zout;
  Zout = 1./x0[Mpfft.numRow()-1];
  fprintf(out_file,"%E %E %E %E %E\n ", freq, Zout.real(), Zout.imag(), x0[Mpfft.numRow()-1].real(), x0[Mpfft.numRow()-1].imag());
  fclose(out_file);
  //}

}

TNT::Vector<complex<double> > sysFormation::MatrixVectorProduct_C(TNT::Vector<complex<double> > RHS)
{
  
  
  TNT::Vector<complex<double> > AP_Vmpfft(RHS.size());
  TNT::Vector<complex<double> > IbX(ElementListX.size());
  TNT::Vector<complex<double> > IbY(ElementListY.size());
  TNT::Vector<complex<double> > IbZ(ElementListZ.size());
  TNT::Vector<complex<double> > IbP(ElementListP.size());
  
  TNT::Vector<complex<double> > Ib(MTpfft.numRow());
  
  matMultVec(Ib,MTpfft,RHS);
  
  for (int ix=0; ix<ElementListX.size(); ix++)
    IbX[ix]=Ib[ix];
  for (int iy=0; iy<ElementListY.size(); iy++)
    IbY[iy]=Ib[iy+ElementListX.size()];
  for (int iz=0; iz<ElementListZ.size(); iz++)
    IbZ[iz]=Ib[iz+ElementListX.size()+ElementListY.size()];
  for (int ip=0; ip<ElementListP.size(); ip++)
    IbP[ip]=Ib[ip+ElementListX.size()+ElementListY.size()+ElementListZ.size()];
  
  
  TNT::Vector<complex<double> > Vb(ElementListX.size()+ElementListY.size()+ElementListZ.size()+ElementListP.size());
  
  TNT::Vector<complex<double> > y_XX(ElementListX.size());
  TNT::Vector<complex<double> > y_YY(ElementListY.size());
  TNT::Vector<complex<double> > y_XZ(ElementListZ.size());
  TNT::Vector<complex<double> > y_YZ(ElementListZ.size());
  TNT::Vector<complex<double> > y_ZZ(ElementListZ.size());
  TNT::Vector<complex<double> > y_P(ElementListP.size());
  
  
  if (ElementListP.size() > 0) {
    pfft_P_H.IEoperator(y_P, outerOperatorList[0], innerOperatorList[0],
			      IbP);
  }
  
  if (ElementListX.size() > 0) {
    pfft_XX.IEoperator(y_XX, outerOperatorList[0], innerOperatorList[0],
		       IbX);
  }
  if (ElementListY.size() > 0) {
    pfft_YY.IEoperator(y_YY, outerOperatorList[0], innerOperatorList[0],
		       IbY);
  }
  if (ElementListX.size() > 0 && ElementListZ.size() > 0) {
    pfft_XZ.IEoperator(y_XZ, outerOperatorListX[0], innerOperatorList[0],
		       IbX);
  }
  if (ElementListY.size() > 0 && ElementListZ.size() > 0) {
    pfft_YZ.IEoperator(y_YZ, outerOperatorListY[0], innerOperatorList[0],
		       IbY);
  }
  if (ElementListZ.size() > 0) {
    pfft_ZZ.IEoperator(y_ZZ, outerOperatorList[0], innerOperatorList[0],
		       IbZ);
  }
  
  for (int i=0; i<ElementListX.size(); i++) {
    Vb[i]=y_XX[i]*IMAG*omega*1e-7;
  }
  for (int i=0; i<ElementListY.size(); i++) {
    Vb[ElementListX.size()+i]=
      y_YY[i]*IMAG*omega*1e-7;
  }
  for (int i=0; i<ElementListZ.size(); i++) {
    Vb[ElementListX.size()+ElementListY.size()+i]
      =(y_ZZ[i]+y_XZ[i]+y_YZ[i])*IMAG*omega*1e-7;
  }
  Vb[ElementListX.size()+ElementListY.size()+ElementListZ.size()]=(0.,0.);
  
  for (int i=0; i<ElementListP.size(); i++) {
    Vb[ElementListX.size()+ElementListY.size()+ElementListZ.size()+i]
      =y_P[i]/(IMAG*omega*4.*M_PI*layers.get_epir_topmostlayer()*EPI0);
  }
  
  matMultVec(AP_Vmpfft,Mpfft,Vb);
  
  TNT::Vector<complex<double> > VR(ElementListX.size()+ElementListY.size()+ElementListZ.size());
  TNT::Vector<complex<double> > Ibr(ElementListX.size()+ElementListY.size()+ElementListZ.size());
  TNT::Vector<complex<double> > VR_PADDED(MTpfft.numRow());
  for (int i=0; i<ElementListX.size()+ElementListY.size()+ElementListZ.size();i++)
    Ibr[i]=Ib[i];
  matMultVec(VR,RMat,Ibr);
  for (int i=0; i<ElementListX.size()+ElementListY.size()+ElementListZ.size();i++)
    VR_PADDED[i]=VR[i];
  for (int i=0; i<ElementListP.size();i++)
    VR_PADDED[i+ElementListX.size()+ElementListY.size()+ElementListZ.size()]=0.;
  
  
  matMultVecAdd(AP_Vmpfft,Mpfft,VR_PADDED);
  
  return AP_Vmpfft;
  
}
/********************************************************
setup_RHS_InitialGuess();
*********************************************************************************/
TNT::Vector<complex<double> > sysFormation::setup_RHS_InitialGuess()
{
  
  TNT::Vector<std::complex<double> > RHSpfft(Mpfft.numRow());
  vector<std::complex<double> > RHSpfft_(Mpfft.numRow());
  vector<std::complex<double> > RHSpfft__(Mpfft.numRow());
  
  for (int i=0; i<Mpfft.numRow()-2; i++)
    //for (int i=0; i<Mpfft.numRow(); i++)
    RHSpfft[i]=(complex<double>(0.,0.));
  
  //  RHSpfft[port_location_]=(complex<double>(1.,0.));
  
  RHSpfft[Mpfft.numRow()-2]=(complex<double>(0.,0.));
  RHSpfft[Mpfft.numRow()-1]=(complex<double>(1.,0.));
  
  /*
    superLU_PreCond.setup(MpreCondMatMT);
    superLU_PreCond.factor();
    cout<<"system has been factored"<<endl;
    for (int i=0; i<RHSpfft.size(); i++) {
    RHSpfft_[i]=RHSpfft[i];
    //cout << RHSpfft_[i] << endl;
    }
    superLU_PreCond.solve(RHSpfft_, RHSpfft__);
    
    for (int i=0; i<RHSpfft.size(); i++) {
    RHSpfft[i]=RHSpfft__[i];
    //	cout << RHSpfft[i] << endl;
    }
    
    cout<<"norm is "<<superLU_PreCond.estimateCondNum_OneNorm()<<endl;
  */
  
  return RHSpfft;	
}





/******************************************************************************
	setup_M this function is used to reorder the rows and vectors of
******************************************************************************/
void sysFormation::setup_M_Elements(void)
{
  const filament<double> *f1;
  panel<double> *p1, *p2;
  //int i,j;
  double f_x, f_y, f_z;
  int numfil_X=0, numfil_Y=0, numfil_Z=0;
  int Tnumfil_X=0, Tnumfil_Y=0, Tnumfil_Z=0, Tnumpanel_H=0, Tnumpanel_V=0;
  //	Mpfft = pfft::SpRowMat<double>(M.numRow()-2, M.numCol()-1);
  //	MTpfft = pfft::SpRowMat<double>(M.numCol()-1, M.numRow()-2);
  //	pfft::SpColMat<double>  MCol(M.numCol()-1, M.numRow());
  cout << M.numCol() << endl;
  Mpfft = pfft::SpRowMat<double>(M.numRow(), M.numCol());
  MTpfft = pfft::SpRowMat<double>(M.numCol(), M.numRow());
  pfft::SpColMat<double>  MCol(M.numCol(), M.numRow());
  
  for (int i=0; i<M.numRow(); i++) {
    for (int j=0; j < M[i].size(); j++) {
      // if (M[i].index(j)<12)
      MCol.insertElement(i,M[i].index(j),M[i].value(j));
      // else if (M[i].index(j)>12)
      // 		MCol.insertElement(i,M[i].index(j)-1,M[i].value(j));
    }}
  
  
  for (f1=condStruc.list_filaments.initialize(); f1!=NULL;
       f1=condStruc.list_filaments.iterate()) {
    f_x= (f1->l_dir()).x();
    f_y= (f1->l_dir()).y();
    f_z= (f1->l_dir()).z();
    if (f_x!=0) Tnumfil_X++;
    else if (f_y!=0) Tnumfil_Y++;
    else if (f_z!=0) Tnumfil_Z++;
    else cout << "ERROR" << endl;
  }
  
  for (p1=condStruc.list_panels.initialize(); p1!=NULL;
       p1=condStruc.list_panels.iterate()) {
    if (p1->direction().x()!=0 || p1->direction().y()!=0) Tnumpanel_H++;
    else if (p1->direction().z()!=0) Tnumpanel_V++;
    else cout << p1->direction().x() << " " << p1->direction().y() << " " << p1->direction().z() << endl;
  }
  
  
  int idx = 0;
  for (f1=condStruc.list_filaments.initialize(); f1!=NULL;
       f1=condStruc.list_filaments.iterate()) {
    f_x= (f1->l_dir()).x();
    f_y= (f1->l_dir()).y();
    f_z= (f1->l_dir()).z();
    pfft::element ele(f1,numfil_X+numfil_Y+numfil_Z,-1,0);
    idx = numfil_X+numfil_Y+numfil_Z;
    if (f_x!=0) {
      for (int R_idx=0; R_idx < MCol[idx].size(); R_idx++) {
      	Mpfft.insertElement(MCol[idx].index(R_idx), numfil_X,  MCol[idx].value(R_idx));
      	MTpfft.insertElement(numfil_X, MCol[idx].index(R_idx), MCol[idx].value(R_idx));
      }
				numfil_X++;
				ElementListX.push_back(ele);
    } else if (f_y!=0) {
      for (int R_idx=0; R_idx < MCol[idx].size(); R_idx++) {
      	Mpfft.insertElement(MCol[idx].index(R_idx), Tnumfil_X+numfil_Y,  MCol[idx].value(R_idx));
      	MTpfft.insertElement(Tnumfil_X+numfil_Y, MCol[idx].index(R_idx), MCol[idx].value(R_idx));
      }
      numfil_Y++;
				ElementListY.push_back(ele);
    } else if (f_z!=0) {
      for (int R_idx=0; R_idx < MCol[idx].size(); R_idx++) {
      	Mpfft.insertElement(MCol[idx].index(R_idx), Tnumfil_X+Tnumfil_Y+numfil_Z,  MCol[idx].value(R_idx));
      	MTpfft.insertElement(Tnumfil_X+Tnumfil_Y+numfil_Z, MCol[idx].index(R_idx), MCol[idx].value(R_idx));
      }
      numfil_Z++;
      ElementListZ.push_back(ele);
    }
  }
  
  cout << "number of source filaments is: " << Tnumfil_X+Tnumfil_Y+Tnumfil_Z << endl;
  
  
  int numPanel_H=0, numPanel_V=0;
  int idxP=0;
  int Tnumfil=Tnumfil_X+Tnumfil_Y+Tnumfil_Z;
  int idxV=Tnumfil+Tnumpanel_H;
  for (p1=condStruc.list_panels.initialize(); p1!=NULL;
       p1=condStruc.list_panels.iterate()) {
    pfft::element ele(p1,0,-2,0);
    idxP = Tnumfil+numPanel_H+numPanel_V;
    if ((p1->direction().x()!=0) || (p1->direction().y()!=0)) {
      for (int R_idx=0; R_idx < MCol[idxP].size(); R_idx++) {
        //if (MCol[idxP].index(R_idx)<M.numRow()-2) {
      	Mpfft.insertElement(MCol[idxP].index(R_idx), Tnumfil+numPanel_H, MCol[idxP].value(R_idx));
      	MTpfft.insertElement(Tnumfil+numPanel_H, MCol[idxP].index(R_idx), MCol[idxP].value(R_idx));
				} //}
      numPanel_H++;
      //	ElementListP.push_back(ele);
      ElementListPH.push_back(ele);
    }
    else if (p1->direction().z()!=0) {
      for (int R_idx=0; R_idx < MCol[idxP].size(); R_idx++) {
        //if (MCol[idxP].index(R_idx)<M.numRow()-2) {
      	Mpfft.insertElement(MCol[idxP].index(R_idx), idxV+numPanel_V,  MCol[idxP].value(R_idx));
      	MTpfft.insertElement(idxV+numPanel_V, MCol[idxP].index(R_idx), MCol[idxP].value(R_idx));
      } //}
      numPanel_V++;
      ElementListPV.push_back(ele);
    }
  }
  for (int i=0; i<ElementListPH.size(); i++)
    ElementListP.push_back(ElementListPH[i]);
  for (int i=0; i<ElementListPV.size(); i++)
    ElementListP.push_back(ElementListPV[i]);
  
  cout << "number of source panels is: " << ElementListP.size() << endl;
  cout << "number of horizontal source panels is: " << numPanel_H << endl;
  cout << "number of vertical source panels is: " << numPanel_V << endl;

  // output_to_file_real("Mpfft.mat", Mpfft);
  // output_to_file_real("MTpfft.mat", MTpfft);
  
  cout << Mpfft.numRow() << endl;
  cout << Mpfft.numCol() << endl;
  
}


/*************************************************************************
* setup pfft this function serves to setup the lements and call
* all the different pfft routines with the appropriate initialization
**************************************************************************/
void sysFormation::setup_pfft_C(const Dipoles& dipoles,
                                const double freq,
				const SimulationType simuType)
  
{
  
  
  setup_freq_info(freq);
  
  
  ofstream fout("error.dat");
  size_t max_size = 1;
  
  
  size_t interpStencilSize =  max_size;
  size_t projectStencilSize = max_size;
  size_t directStencilSize = 2*max_size;
  cout << endl << "\t Calculate matrix vector product indirectly using pfft" << endl;
  
  outerOperatorList.push_back(pfft::D0_D0);
  innerOperatorList.push_back(pfft::D0_D0);
  outerOperatorListX.push_back(pfft::D_DX);
  innerOperatorListX.push_back(pfft::D_DX);
  outerOperatorListY.push_back(pfft::D_DY);
  innerOperatorListY.push_back(pfft::D_DY);
  
  
  if (ElementListX.size() > 0) {
	  calcA_XX=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0, dipole_flag,
				   dipoles, pfft::GXXc,simuType);
	  
	  DynamicGreenFunc kernel_XX(k0, dipoles, pfft::GXXc);
	  kernel_XX.set_shift(get_shift(ElementListX));
	  pfft_XX.PfftConstruct(ElementListX, ElementListX, kernel_XX, calcA_XX,
				directStencilSize, projectStencilSize, interpStencilSize, pfft::TOEPLITZplusHANKEL);
  }
  
  if (ElementListY.size() > 0) {
    calcA_YY=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0, dipole_flag,
			     dipoles, pfft::GXXc,simuType);
    
    DynamicGreenFunc kernel_YY(k0, dipoles, pfft::GXXc);
    kernel_YY.set_shift(get_shift(ElementListY));
    pfft_YY.PfftConstruct(ElementListY, ElementListY, kernel_YY, calcA_YY,
			  directStencilSize, projectStencilSize, interpStencilSize, pfft::TOEPLITZplusHANKEL);
  }
  
	if (ElementListX.size() > 0 && ElementListZ.size() > 0) {
	  calcA_XZ=CalcAfullwave3D(outerOperatorListX, innerOperatorList, k0, dipole_flag, dipoles,
				   pfft::GXZc,simuType);
	  
	  DynamicGreenFunc kernel_XZ(k0, dipoles, pfft::GXZc);
	  kernel_XZ.set_shift(get_shift(ElementListX));
	  pfft_XZ.PfftConstruct(ElementListX, ElementListZ, kernel_XZ, calcA_XZ,
				directStencilSize, projectStencilSize, interpStencilSize, pfft::HANKEL);
	}
	
	if (ElementListY.size() > 0 && ElementListZ.size() > 0) {
	  calcA_YZ=CalcAfullwave3D(outerOperatorListY, innerOperatorList, k0, dipole_flag,
				   dipoles, pfft::GXZc,simuType);
	  
	  DynamicGreenFunc kernel_YZ(k0, dipoles, pfft::GXZc);
	  kernel_YZ.set_shift(get_shift(ElementListY));
	  pfft_YZ.PfftConstruct(ElementListY, ElementListZ, kernel_YZ, calcA_YZ,
				directStencilSize, projectStencilSize, interpStencilSize, pfft::HANKEL);
	}
	if (ElementListZ.size() > 0) {
	  calcA_ZZ=CalcAfullwave3D(outerOperatorList, innerOperatorList, k0, dipole_flag,
				   dipoles, pfft::GZZ,simuType);
	  DynamicGreenFunc kernel_ZZ(k0, dipoles, pfft::GZZ);
	  kernel_ZZ.set_shift(get_shift(ElementListZ));
	  pfft_ZZ.PfftConstruct(ElementListZ, ElementListZ, kernel_ZZ, calcA_ZZ,
				directStencilSize, projectStencilSize, interpStencilSize, pfft::TOEPLITZplusHANKEL);
	}
	
	if (ElementListP.size() > 0) {
	  
	  if (simuType==FULLWAVE) {
	    calcP_VERT=CalcPfullwave3D(outerOperatorList, innerOperatorList, k0,
				       dipole_flag, dipoles, pfft::GZZ_phi, simuType);
	    kernel_P=DynamicGreenFunc(k0, dipoles, pfft::GZZ_phi);
	  } else if (simuType==EMQS) {
	    calcP_VERT=CalcPfullwave3D(outerOperatorList, innerOperatorList, k0,
				       dipole_flag, dipoles, pfft::GZZ_phi, simuType);
		 kernel_P=DynamicGreenFunc(0., dipoles, pfft::GZZ_phi);
	  }
	  kernel_P.set_shift(get_shift(ElementListP));
	  pfft_P_H.PfftConstruct(ElementListP, ElementListP, kernel_P, calcP_VERT,
				 directStencilSize, projectStencilSize, interpStencilSize, pfft::TOEPLITZplusHANKEL);
	}
	
	setup_PRECONDITIONER();
	
}



void sysFormation::setup_PRECONDITIONER(void)
{
  
  TNT::Vector<complex<double> > ans(Mpfft.numCol());
  //pfft::SpVec<std::complex<double> > ans(Mpfft.numCol());
  MpreCondMatMT=pfft::SpColMat<std::complex<double> >(Mpfft.numRow(),Mpfft.numRow());
  
  for (int i = 0; i < Mpfft.numRow(); i++)
    ans[i] = 0.;
  
  for(size_t R_idx=0; R_idx < Mpfft.numRow(); R_idx++)
    {
      if ((((int)R_idx) % 100)==0)
	cout << "finished row # = " << R_idx  << endl;
      
      for (size_t i = 0; i < Mpfft[R_idx].size(); i++)
	{
	  size_t row = Mpfft[R_idx].index(i);
	  for (size_t col = 0; col < MTpfft[row].size(); col++)
	    {
	      if (row < ElementListX.size())
		ans[MTpfft[row].index(col)] += Mpfft[R_idx].value(i) * MTpfft[row].value(col) * (pfft_XX.getPreCondMat())[row] * IMAG*omega*1e-7;
	      else if (row < ElementListX.size()+ElementListY.size())
		ans[MTpfft[row].index(col)] += Mpfft[R_idx].value(i) * MTpfft[row].value(col) * (pfft_YY.getPreCondMat())[row-ElementListX.size()] * IMAG*omega*1e-7;
	      else if (row < ElementListX.size()+ElementListY.size()+ElementListZ.size())
		ans[MTpfft[row].index(col)] += Mpfft[R_idx].value(i) * MTpfft[row].value(col) * (pfft_ZZ.getPreCondMat())[row-ElementListX.size()-ElementListY.size()] * IMAG*omega*1e-7;
	      else if (row < ElementListX.size()+ElementListY.size()+ElementListZ.size()+ElementListP.size())
		ans[MTpfft[row].index(col)] += Mpfft[R_idx].value(i) * MTpfft[row].value(col) * (pfft_P_H.getPreCondMat())[row-ElementListX.size()-ElementListY.size()-ElementListZ.size()] / 
		  (IMAG*omega*4.*M_PI*layers.get_epir_topmostlayer()*EPI0);
	      else cout << "ERROR" << endl;
	      //MpreCondMatMT.insertElement(R_idx, MTpfft[row].index(col), ans[MTpfft[row].index(col)]);
	    }
	}
      
      for (int i = 0; i < Mpfft.numRow(); i++) {
	if (ans[i] !=0.) MpreCondMatMT.insertElement(R_idx, i, ans[i]);
	ans[i]=0.;
      }
      
      
    }
}


/////////////////////////
double sysFormation::get_shift(vector<pfft::element> ElementList)
{
double zmin=ElementList[0].zmin();
	for (int i=1; i<ElementList.size(); i++)
		{
		zmin=min(zmin,ElementList[i].zmin());
		}
return zmin;
}
